/* 
  $Id: getCoverage.c,v 1.1.1.1 2005/08/04 14:32:29 mcschatz Exp $

  main() routine of getCoverage

  Computes a fragment or/and global tiling information given prefix.contig and prefix.qual files.
   
  initial design by Pawel Gajer
 */

#include <unistd.h>
#include "getCoverage.h"


int getCoverage(const char * prefix, const char * asmbl_id,
                int * ungapped_pos, int * gapped_pos, int radius,
                int * ungapped_end, int * gapped_end, int coordinateOffset,
                enum DisplayMode displayMode, int print_offset, int doBasesCoverage, 
                int doCalcConsensus, int allowGaps, const char * output_prefix, 
                int doSplit, int showReadId, int doGapped, char allowMissing,
                int doGapQV, const char * seqname_range, char useIndex,
                char assumeCircular)
{
  char qualFile   [FILENAME_MAX];
  char contigFile [FILENAME_MAX];

  FILE *f;
  int numOfSeqs;
  int numOfQVSeqs = 0;
  Offset *qvIndex = NULL;

  char c;
  int count = 1; 
  int foundContig = 0;
  int doAll = !strcmp(asmbl_id, "all") || doSplit;

  struct CONTIG_SEQUENCE *contig;
  struct CONTIG_SEQUENCE **bcseq;
  tilingData td;

  off_t contigoffset;

  Coordinates coords;
  coords.coordinateOffset = coordinateOffset;

  if (displayMode != VERTICAL && displayMode != HORIZONTAL)
  {
    // index the qual values
    strcpy(qualFile, prefix);    
    strcat(qualFile, ".qual");

    f = fopen(qualFile, "r");

    if (!f && allowMissing == -1)
    {
      printError("getCoverage", DEBUG_ERROR,
                 "ERROR: Couldn't open %s for reading", qualFile);

      return 1; // Missing Qual File
    }
    
    if (f)
    {
      setQualFile(f);
      logMsg(DEBUG_MEDIUM, "getCoverage",
             "Creating index of offsets of seq_ids in %s", qualFile); 

      indexSeqIDs(f, qualFile, &qvIndex, &numOfQVSeqs, useIndex);

      logMsg(DEBUG_MEDIUM, "getCoverage", "done");
    }
    else
    {
      logMsg(DEBUG_MEDIUM, "getCoverage",
             "Couldn't open quality values, using dummy values");
    }
  }
  
  logMsg(DEBUG_LOW, "getCoverage", "ENTERING THE MAIN LOOP");
  
  // open .contig file
  strcpy(contigFile, prefix);
  strcat(contigFile, ".contig");
  f = fOpen(contigFile, "r");

  contigoffset = ftello(f);

  do
  {
    logMsg(DEBUG_MEDIUM, "getCoverage", "%d: ", count);
    count++;
    
    fseeko(f, contigoffset, SEEK_SET);
    readContig(f, &contig, &numOfSeqs);

    logMsg(DEBUG_MEDIUM, "getCoverage", 
          "Processing assembly %s with %d reads",
          contig->id, numOfSeqs); 

    foundContig = (asmbl_id[0]=='\0') || // none specified
                  !strcmp(asmbl_id, contig->id); //found the right one

    if (foundContig || doAll)
    {
      convertCoordinates(contig, ungapped_pos, gapped_pos, 
                         radius, ungapped_end, gapped_end,
                         seqname_range, &coords);

      logMsg(DEBUG_LOW, "getCoverage", 
             "Loading sequences for contig %s", contig->id);

      // Only load the sequences that are of interest (possibly all)
      c = readSeqs(&numOfSeqs, &bcseq, contig, 
                   &coords, seqname_range, assumeCircular,
                   displayMode == TCOV);

      contigoffset = ftello(f); // remember where the next asmbl_id begins

      if (displayMode != VERTICAL && displayMode != HORIZONTAL)
      {
        logMsg(DEBUG_MEDIUM, "getCoverage", "Parsing %s ... ", qualFile);

        parseQual(qvIndex, numOfQVSeqs, 
                  numOfSeqs, bcseq, contig->id,
                  allowMissing, doGapQV);

        logMsg(DEBUG_MEDIUM, "getCoverage", "done.");  
      }

      logMsg(DEBUG_MEDIUM, "getCoverage",
             "Aligning sequences to the consensus sequence ... "); 
      alignSeqs(bcseq, &td, numOfSeqs, contig->len);

      logMsg(DEBUG_MEDIUM, "getCoverage", "done.");  

      switch (displayMode)
      {
        case TCOV:
        case TCOV_UNSTABLE:
          logMsg(DEBUG_MEDIUM, "getCoverage", "Printing TotalCov");
          printTotalCov(contig, bcseq, &td, allowGaps, &coords, output_prefix,
                        prefix, doAll && !doSplit, print_offset,
                        showReadId, doGapped);
          break;

        case VERTICAL:
          logMsg(DEBUG_MEDIUM, "getCoverage", "Printing Vertical");
          printVertical(contig, bcseq, &td, &coords, output_prefix, prefix); 
          break;

        case STATS:
          logMsg(DEBUG_MEDIUM, "getCoverage", "Printing Stats");
          printStats(contig, bcseq, &td, allowGaps, &coords,
                     output_prefix, prefix, doAll && !doSplit, 
                     doBasesCoverage, doCalcConsensus);
          break;

        case HORIZONTAL:
          logMsg(DEBUG_MEDIUM, "getCoverage", "Printing Horizontal");
          printHorizontal(contig, bcseq, &td, doGapped, &coords, output_prefix);
          break;

        case XML:
          logMsg(DEBUG_MEDIUM, "getCoverage", "Printing XML");
          printXML(contig, bcseq, &td, allowGaps, &coords, output_prefix);
          break;

        default:
          logMsg(DEBUG_MEDIUM, "getCoverage", "Printing Coverage");
          printCov(contig, bcseq, &td, allowGaps, &coords, output_prefix);
      };

      freeCovData(&td);
    }
    else
    {
      coords.start = coords.ugInd = coords.end = -1;

      // Just advance to next contig record
      c = readSeqs(&numOfSeqs, &bcseq, contig, &coords,
                   NULL, assumeCircular, 0);
      contigoffset = ftello(f);
    }

    logMsg(DEBUG_MEDIUM, "getCoverage", 
           "done processing asmbl_id %s", contig->id); 

    freeSeqData(bcseq, numOfSeqs);
    freeContigData(contig);

    pool_flush();

    if (foundContig && !doAll) break;
  } while (c != EOF);

  fclose(f);
  freeOffset(qvIndex, numOfQVSeqs);
  closeQualFile();

  if (!foundContig && !doAll && asmbl_id[0] != '\0')
  {
    printError("getCoverage", DEBUG_ERROR,
               "ERROR: asmbl_id %s not found", asmbl_id);
    return 1;
  }

  logMsg(DEBUG_MEDIUM, "getCoverage", "END OF MAIN LOOP");
  
  return 0;
}
