#include "SliceAugmenter.hh"
#include "FileSystem.hh"

SliceAugmenter::SliceAugmenter() : RequestManager("SliceAugmenter")
{
  m_log.debugL("Contructed SliceAugmenter");
  m_log.setMessageLevel(MESSAGE_LEVEL_DEBUG_MEDIUM);
}

SliceAugmenter::~SliceAugmenter()
{
  m_log.debugL("Destructed SliceAugmenter");
}

void SliceAugmenter::addSliceRange(long & index, long & gindex)
{
  // we've seen all the reads, now load the seq and qual
  processPrefix();

  m_out << "  </ReadCollection>" << endl;

  m_out << "  <SliceRange Index=\"" << index
        << "\" Gindex=\""           << gindex
        << "\">" << endl;

  index = gindex = 0L;
}

void SliceAugmenter::addSlice(SliceData * slice)
{
  RequestManager::addSlice(slice);

  slice->printXML(m_out);
  delete slice;
}


void SliceAugmenter::addReadData(ReadData * readData)
{
  const string & seqname = readData->getSeqname();
  m_readMap[seqname] = readData;
  m_readList.push_back(readData);
}

void SliceAugmenter::setAugmentedData(ReadData * readData,
                                      TrimmedSequence * seqdata)
{
  if (seqdata)
  {
    const string & nuc = seqdata->getNuc();
    const string & qual = seqdata->getQual();
    const string & seqname = readData->getSeqname();

    if (nuc.length() && qual.length())
    {
      string scratch;

      long seqleft = readData->getClrleft();
      long seqright = readData->getClrright();

      try
      {
        scratch = nuc.substr(0, seqleft-1);
        readData->addNucData(scratch.c_str(), true);
      } 
      catch (...)
      {
        cerr << "Unexpected error in addNucData for " << seqname
             << ", true" << endl;
      }

      try
      {
        long endpos = ((seqleft-1) * 3) - 1;
        scratch = qual.substr(0, endpos);
        readData->addQualData(scratch.c_str(), true);
      }
      catch (...)
      {
        cerr << "Unexpected error in addQualData for " << seqname
             << ", true" << endl;
      }

      try
      {
        scratch = nuc.substr(seqright, nuc.length());
        readData->addNucData(scratch.c_str(), false);
      }
      catch (...)
      {
        cerr << "Unexpected error in addNucData for " << seqname 
             << ", false" << endl;
      }

      long endpos = qual.length();
      long startpos = seqright*3;

      if ((endpos-startpos+1) > 0)
      {
        try
        {
          scratch = qual.substr(startpos, endpos-startpos+1);
          readData->addQualData(scratch.c_str(), false);
        }
        catch (...)
        {
          cerr << "Unexpected error in addQualData for " << seqname
               << ", false" << endl;
          cerr << "startpos is " << startpos
               << " endpos is " << endpos
               << "count is " << endpos - startpos + 1
               << endl;
        }
      }
    }
    else
    {
      cerr << "No data to augment for " << seqname << endl;
    }
  }
}

void SliceAugmenter::addRequest(ConsensusRequest * request)
{
  m_requestData = request;

  printXMLHeader(request);

  m_out << "  <ReadCollection>" << endl;
}

int SliceAugmenter::printResults()
{
  m_out << "  </SliceRange>" << endl;
  m_out << "</Request>"      << endl;

  return m_errorCount;
}

void SliceAugmenter::loadIndex(ifstream & file, bool isQual)
{
  string seqname;
  int qlen;
  long long offset;

  if (isQual)
  {
    while (file >> seqname >> qlen >> offset)
    {
      //cerr << "q:" << seqname << " " << qlen << " " << offset << endl;
      
      if (m_readMap.find(seqname) != m_readMap.end())
      {
        m_qualidx[seqname] = offset-1;
      }
    }
  }
  else
  {
    while (file >> seqname >> offset)
    {
      //cerr << "s:" << seqname << " " << offset << endl;
      
      if (m_readMap.find(seqname) != m_readMap.end())
      {
        m_seqidx[seqname] = offset;
      }
    }
  }
}

TrimmedSequence * SliceAugmenter::handleBuffer(char * buffer, TrimmedSequence * currentSequence, bool isQual)
{
  if (buffer[0] == '>')
  {
    char * whitespace = strchr(buffer, ' ');
    if (whitespace) { *whitespace = '\0'; }
    string seqname = buffer+1;

    bool loadSequence = (m_readMap.find(seqname) != m_readMap.end());
    currentSequence = NULL;

    if (loadSequence)
    {
      //cerr << "Loading " << seqname << " " << isQual << endl;
      if (isQual)
      {
        currentSequence = new TrimmedSequence(seqname);
        m_seqMap[seqname] = currentSequence;
      }
      else
      {
        currentSequence = m_seqMap[seqname];
        if (!currentSequence)
        {
          throw Exception ("Found seq without qual values",
                               "SliceAugmenter::parseFile");
        }
      }
    }
  }
  else if (currentSequence)
  {
    currentSequence->addData(buffer, isQual);
  }

  return currentSequence;
}

void SliceAugmenter::parseFile(ifstream & file, bool isQual)
{
  TrimmedSequence * currentSequence = NULL;

  char buffer[2048];

  if ((isQual && !m_qualidx.empty()) || (!isQual && !m_seqidx.empty()))
  {
    OffsetMap::iterator oi;
    ReadList::iterator ri;

    for (ri = m_readList.begin(); ri != m_readList.end(); ri++)
    {
      //cerr << "ri: " << (*ri)->getSeqname() << " ";
      if (isQual)
      {
        oi = m_qualidx.find((*ri)->getSeqname());
        if (oi == m_qualidx.end())
        {
          throw Exception("Didn't find " + (*ri)->getSeqname() + " in the qual idx", "parseFile");
        }
      }
      else
      {
        oi = m_seqidx.find((*ri)->getSeqname());
        if (oi == m_seqidx.end())
        {
          throw Exception("Didn't find " + (*ri)->getSeqname() + " in the seq idx", "parseFile");
        }
      }

      file.clear();
      file.seekg(oi->second);

      // Get the >seqname
      file.getline(buffer, sizeof(buffer));
      currentSequence = handleBuffer(buffer, currentSequence, isQual);

      while (file && file.getline(buffer,sizeof(buffer)) && buffer[0] != '>')
      {
        currentSequence = handleBuffer(buffer, currentSequence, isQual);
      }
    }
    //cerr << "Done" << endl;
  }
  else // No index, just try the whole file
  {
    while(file)
    {
      file.getline(buffer, sizeof(buffer));
      currentSequence = handleBuffer(buffer, currentSequence, isQual);
    }
  }
}





void SliceAugmenter::processPrefix()
{
  try 
  {
    ifstream * qualidx = FileSystem::openInputFile(m_prefix + ".qual.idx");
    loadIndex(*qualidx, true);
    delete qualidx;
  }
  catch (const Exception & e)
  {
    //cerr << "No qual idx" << endl;
  }

  ifstream * qualFile = FileSystem::openInputFile(m_prefix + ".qual");
  parseFile(*qualFile, true);
  delete qualFile;

  try 
  {
    ifstream * seqidx = FileSystem::openInputFile(m_prefix + ".seq.idx");
    loadIndex(*seqidx, false);
    delete seqidx;
  }
  catch (const Exception & e)
  {
    //cerr << "No seq idx" << endl;
  }

  ifstream * seqFile = FileSystem::openInputFile(m_prefix + ".seq");
  parseFile(*seqFile, false);
  delete seqFile;

  ReadList::iterator ri;

  for (ri =  m_readList.begin();
       ri != m_readList.end();
       ri++)
  {
    const string & seqname = (*ri)->getSeqname();
    TrimmedSequence * ts = m_seqMap[seqname];
    setAugmentedData(*ri, ts);
    delete(ts);
 
    (*ri)->printXML(m_out);
  }
}

void SliceAugmenter::setPrefix(const string & prefix)
{
  m_prefix = prefix;
}
