#include "Edit.hh"
#include <stdio.h>

Edit::Edit(const string & contigId,
           long gindex,
           long index,
           long seqPos,
           const string & seqName,
           char origBase,
           char editBase,
           int rc,
           int qv,
           LogCategory log,
           int editFilterLevel)
  : m_log(log)
{
  m_contigId = contigId;
  m_gindex   = gindex;
  m_index    = index;
  m_seqPos   = seqPos;
  m_seqName  = seqName;
  m_origBase = origBase;
  m_editBase = editBase;
  m_rc       = rc;
  m_qv       = qv;

  m_editFilterLevel = editFilterLevel;

  if (m_origBase == '*') { m_origBase = '-'; }
  if (m_editBase == '*') { m_editBase = '-'; }

  m_verifyIndex = false;

  m_log.setMessageLevel(MESSAGE_LEVEL_DEBUG_INFORMATION);
}

Edit::~Edit()
{ }

void Edit::printXML(ostream & out, const string & prefix)
{
  out << prefix << m_contigId 
      << " "    << m_gindex
      << " "    << m_index
      << " "    << m_seqPos
      << " "    << m_seqName
      << " "    << m_origBase
      << " "    << m_editBase
      << " "    << m_rc
      << " "    << m_qv
      << endl;

  /*
  out << prefix 
      << "<Edit contigId=\"" << m_contigId
      << "\" index=\"" << m_index
      << "\" seqPos=\""    << m_seqPos
      << "\" seqName=\""   << m_seqName
      << "\" origBase=\""   << m_origBase
      << "\" editBase=\""   << m_editBase
      << "\" rc=\""        << m_rc
      << "\" />" << endl;
  */
}

// Applies an edit to a slice, validates the expected base is found in the
// slice and that the proper read is found
int Edit::apply(SliceData * slice, long seqpos)
{
  bool foundSeq = false;

  unsigned int depth = slice->getDepth();
  string bases = slice->getNucData();

  string qv;
  char buffer[16];

  // Walk the slice to apply the edit
  for (unsigned int i = 0; i < depth; i++)
  {
    int tmpqv = slice->getQualVal(i);

    if (i) { qv += " "; }

    const libSlice::ReadData * rd =  slice->getRead(i);

    if (rd->getSeqname() == m_seqName)
    {
      // Found the right sequence
      foundSeq = true;
      char sliceBase = bases[i];
      
      if (m_rc) { sliceBase = SliceData::getBaseCompliment(sliceBase); }

      int applyEdit = 1;

      if (m_origBase != '-' || m_editBase != '-')
      {
        if (sliceBase != m_origBase)
        {
          cerr << "Expected " << m_editBase << " but found " << sliceBase
                << " at " << m_seqName << "(" << rd->getId() <<  "):" 
                << m_gindex << endl;

          m_log << "Expected " << m_editBase << " but found " << sliceBase
                << " at " << m_seqName << "(" << rd->getId() <<  "):" 
                << m_gindex << endl;
        }

        int modifiedqv = (m_origBase == '-') ? m_qv : tmpqv;
        int editLevel = 0;

        if      (modifiedqv < 15) { editLevel = 1; m_log << "HARMLESS: ";  }
        else if (modifiedqv < 25) { editLevel = 2; m_log << "SAFE: ";      }
        else if (modifiedqv < 35) { editLevel = 3; m_log << "POSSIBLE: ";  }
        else if (modifiedqv < 45) { editLevel = 4; m_log << "UNLIKELY: ";  }
        else                      { editLevel = 5; m_log << "DANGEROUS: "; }

        m_log << m_seqName << "(" << rd->getId() << "):" 
              << m_gindex << "{" << slice->getIndex() << "}" << " "
              << m_origBase << "(" << tmpqv << ")" << "=>" 
              << m_editBase << "(" << m_qv << ")";

        if (m_editFilterLevel)
        {
          if (editLevel > m_editFilterLevel)
          {
            applyEdit = 0;
            m_log << " FILTERED";
          }
          else
          {
            applyEdit = 1;
            m_log << " APPLIED";
          }
        }

        m_log << "\n";
      }

      if (applyEdit)
      {
        tmpqv = m_qv;
        bases[i] = (m_rc) ? SliceData::getBaseCompliment(m_editBase) 
                          : m_editBase;
      }
    }

    snprintf(buffer, sizeof(buffer), "%d", tmpqv);
    qv += buffer;
  }

  // Verify seq was found
  if (!foundSeq)
  {
    cerr << "Didn't find sequence " <<  m_seqName 
         << " at " << m_gindex << endl;

    m_log << "Didn't find sequence " <<  m_seqName 
          << " at " << m_gindex << endl;
  }

  // Verify the position of the edit
  if (m_verifyIndex)
  {
    long sliceIndex = slice->getIndex();
    if (m_index != sliceIndex && (m_editBase != '-' || m_origBase != '-'))
    {
      cerr << "asmPos of edit ("  << sliceIndex
           << ") was unexpected (" << m_index 
           << ") for read " << m_seqName
           << " " << m_origBase << "->" << m_editBase 
           << " at " << m_gindex << endl;

      m_log << "asmPos of edit ("  << sliceIndex
            << ") was unexpected (" << m_index 
            << ") for read " << m_seqName
            << " " << m_origBase << "->" << m_editBase 
            << " at " << m_gindex << endl;
    }

    if (!m_rc) { seqpos-=2; }

    // Adjust seqpos for gap to nongap case
    if (m_origBase == '-' && m_editBase != '-') 
    { 
      if (m_rc) { seqpos--; }
    }

#if 0

    if (m_origBase != '-' || m_editBase != '-')
    {
      if (seqpos != m_seqPos)
      {
        cerr << "Seqpos for slice (" << seqpos 
             << ") was unexpected (" << m_seqPos;
      }
      else
      {
        cerr << "MATCH: Seqpos for slice (" << seqpos;
      }

      cerr << ") for read " << m_seqName
           << " at " << m_gindex
           << " " << m_origBase << ">" << m_editBase;

      if (m_rc) { cerr << " R" << endl; }
      else      { cerr << " F" << endl; }
    }
#endif
  }

  slice->resetNuc();
  slice->addNuc(bases.c_str());

  slice->resetQualVal();
  slice->addQualVal(qv.c_str());

  return (m_editBase == '-' && m_origBase == '-') ? 1 : 0;
}

const string & Edit::getSeqname() const
{
  return m_seqName;
}

void Edit::setVerifyIndex(bool verifyIndex)
{
  m_verifyIndex = verifyIndex;
}

int Edit::seqOffset()
{
  int retval = 0;

  if (m_origBase == '-') 
  {
    if (m_editBase != '-')
    {
      retval = 1;
    }
  }
  else if (m_editBase == '-')
  {
    retval = -1;
  }

  return retval;
}
           
