#include "SliceSAX2Handler.hh"
#include <xercesc/util/PlatformUtils.hpp>
#include "SliceXMLStrings.hh"

//! Constructor initialize members to prepare for accepting XML input
SliceSAX2Handler::SliceSAX2Handler(RequestManager * manager, 
                                   LogCategory & log,
                                   bool doLogging) 
  : m_log(log)
{
  if (!manager)
  {
    throw Exception("Request requires a manager", "Request::Request");
  }

  m_manager = manager;
  m_manager->addReadManager(&m_readManager);

  m_location = IN_UNKNOWN;

  m_attrCount = 0;
  m_elementCount = 0;
  m_characterCount = 0;

  m_readCount = 0;
  m_sliceCount = 0;
  m_sliceRangeCount = 0;

  m_currentSlice = NULL;
  m_currentRead = NULL;

  m_sliceRangeIndex = 0;
  m_sliceRangeGindex = 0;

  m_log.debugL("Created SliceSAX2Handler");
  m_log.setMessageLevel(MESSAGE_LEVEL_DEBUG_HIGH);
  m_doLogging = doLogging && 
                  (m_log.getMessageLevel() <= m_log.getFilterLevel());

  // Mark the start time
  m_startMillis = XMLPlatformUtils::getCurrentMillis();
}

//! Decontructor deletes all allocated memory
SliceSAX2Handler::~SliceSAX2Handler()
{ 
  m_log.debugH("Destroyed SliceSAX2Handler");
}

// ---------------------------------------------------------------------------
//  SliceSAX2Handler: Overrides of the SAX ErrorHandler interface
// ---------------------------------------------------------------------------

//! Called by XML API when an error is encounter in parsing
void SliceSAX2Handler::error(const SAXParseException& e)
{
  throw e;
}

//! Called by XML API when a fatal error is encounter in parsing
void SliceSAX2Handler::fatalError(const SAXParseException& e)
{
  throw e;
}

//! Called by XML API when a warning condition is encounter in parsing
void SliceSAX2Handler::warning(const SAXParseException& e)
{
  throw e;
}

void SliceSAX2Handler::printStats(ostream & out, const string & infile)
{
  unsigned long endMillis = XMLPlatformUtils::getCurrentMillis();
  unsigned long duration = endMillis - m_startMillis;

  int spaceCount = infile.length() + 2;
  string prefix;
  while (spaceCount)
  {
    prefix += " ";
    spaceCount--;
  }

  out << endl 
      << infile << ": " 
      << duration             << " ms, " 
      << m_sliceRangeCount    << " SliceRange(s), "
      << m_sliceCount         << " slice(s), "
      << m_readManager.size() << " read(s)" 
      << endl;

  out << prefix
      << m_elementCount       << " elems, "
      << m_attrCount          << " attrs, "
      << m_characterCount     << " chars" 
      << endl;

  m_manager->printStats(out, prefix);
}


// ---------------------------------------------------------------------------
//  SliceSAX2Handler: Implementation of the SAX DocumentHandler interface
// ---------------------------------------------------------------------------
//! Parses the incoming XML and sends the data to the ConsensusRequest
/*! A list of possible valid XML strings is built at compile time, and
 *  at runtime the incoming element is matched against that list to determine
 *  which element is starting. This way a fixed number of strings need to go
 *  through the expensive XMLString::transcode function.
 *
 *  Once the element is started, its attributes are parsed in the same manner
 *  and then a new object can be created if needed. Transforming the attributes
 *  is unfortunately unavoidable unless the request objects were all converted
 *  to be dependent on XMLCh strings.
 */
void SliceSAX2Handler::startElement(const XMLCh* const uri, 
                                    const XMLCh* const localname,
                                    const XMLCh* const qname,
                                    const Attributes& attrs)
{
  unsigned int attrsLen = attrs.getLength();
  m_attrCount += attrsLen;

  m_elementCount++;

  if (m_doLogging)
  {
    m_log << "startElement: " << StrX(localname) << endl;
    for (unsigned int i = 0; i < attrsLen; i++)
    {
      m_log << "   attribute: \"" << StrX(attrs.getLocalName(i))
            << "\"=\""            << StrX(attrs.getValue(i))
            << "\"" << endl;
    }
  }

  if (!XMLString::compareString(localname, SliceXMLString::sSlice))
  {
    // <Slice> Information (Most common element first)
    
    char * index = NULL;
    char * gindex = NULL;
    char * existing = NULL;

    for (unsigned int i = 0; i < attrsLen; i++)
    {
      const XMLCh * attrname = attrs.getLocalName(i);
      char * attrval = XMLString::transcode(attrs.getValue(i));

      if (!XMLString::compareString(attrname, SliceXMLString::sIndex))
      {
        index = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sGindex))
      {
        gindex = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sExisting))
      {
        existing = attrval;
      }
    }

    handleSliceData(index, gindex, existing);

    if (index)    delete [] index;
    if (gindex)   delete [] gindex;
    if (existing) delete [] existing;
  }
  else if (!XMLString::compareString(localname, SliceXMLString::sNuc))
  {
    // Mark that <Nuc> is the current element for character data
    if (m_location == IN_LEFT || m_location == IN_LEFT_QUAL)
    {
      m_location = IN_LEFT_NUC;
    }
    else if (m_location == IN_RIGHT || m_location == IN_RIGHT_QUAL)
    {
      m_location = IN_RIGHT_NUC;
    }
    else
    {
      m_location = IN_NUC;
    }
  }
  else if (!XMLString::compareString(localname, SliceXMLString::sQualval))
  {
    // Mark that <Qualval> is the current element for character data
    if (m_location == IN_LEFT || m_location == IN_LEFT_NUC)
    {
      m_location = IN_LEFT_QUAL;
    }
    else if (m_location == IN_RIGHT || m_location == IN_RIGHT_NUC)
    {
      m_location = IN_RIGHT_QUAL;
    }
    else
    {
      m_location = IN_QUALVAL;
    }
  }
  else if (!XMLString::compareString(localname, SliceXMLString::sReadID))
  {
    // Mark that <Read> is the current element for character data
    m_location = IN_READID;
  }
  else if (!XMLString::compareString(localname, SliceXMLString::sRead))
  {
    // <Read> Information (next most common element)
    
    char * id = NULL;
    char * dir = NULL;
    char * chem = NULL;
    char * seqname = NULL;

    for (unsigned int i = 0; i < attrsLen; i++)
    {
      const XMLCh * attrname = attrs.getLocalName(i);
      char * attrval = XMLString::transcode(attrs.getValue(i));

      if (!XMLString::compareString(attrname, SliceXMLString::sId))
      {
        id = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sSeqname))
      {
        seqname = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sDir))
      {
        dir = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sChem))
      {
        chem = attrval;
      }
    }

    handleReadData(id, seqname, dir, chem);

    if (id)       delete [] id;
    if (dir)      delete [] dir;
    if (chem)     delete [] chem;
    if (seqname)  delete [] seqname;
  }
  else if (!XMLString::compareString(localname, SliceXMLString::sClr))
  {
    // <Clr> Information
    
    char * left = NULL;
    char * right = NULL;

    for (unsigned int i = 0; i < attrsLen; i++)
    {
      const XMLCh * attrname = attrs.getLocalName(i);
      char * attrval = XMLString::transcode(attrs.getValue(i));

      if (!XMLString::compareString(attrname, SliceXMLString::sLeft))
      {
        left = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sRight))
      {
        right = attrval;
      }
    }

    handleClrData(left, right);

    if (left)       delete [] right;
    if (right)      delete [] left;
  }
  else if (!XMLString::compareString(localname, SliceXMLString::sTrimmedSeq))
  {
    // <TrimmedSeq> Information
    
    char * position = NULL;

    for (unsigned int i = 0; i < attrsLen; i++)
    {
      const XMLCh * attrname = attrs.getLocalName(i);
      char * attrval = XMLString::transcode(attrs.getValue(i));

      if (!XMLString::compareString(attrname, SliceXMLString::sposition))
      {
        position = attrval;
      }
    }

    handleTrimmedSeq(position);

    if (position) delete [] position;
  }
  else if (!XMLString::compareString(localname, SliceXMLString::sSliceRange))
  {
    // <SliceRange> Information

    char * index = NULL;
    char * gindex = NULL;

    for (unsigned int i = 0; i < attrsLen; i++)
    {
      const XMLCh * attrname = attrs.getLocalName(i);
      char * attrval = XMLString::transcode(attrs.getValue(i));

      if (!XMLString::compareString(attrname, SliceXMLString::sIndex))
      {
        index = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sGindex))
      {
        gindex = attrval;
      }
    }

    handleSliceRange(index, gindex);

    if (index)    delete [] index;
    if (gindex)   delete [] gindex;
  }
  else if (!XMLString::compareString(localname, SliceXMLString::sSimpleOption))
  {
    // <SimpleOption> Information
    
    char * name = NULL;
    char * value = NULL;

    for (unsigned int i = 0; i < attrsLen; i++)
    {
      const XMLCh * attrname = attrs.getLocalName(i);
      char * attrval = XMLString::transcode(attrs.getValue(i));

      if (!XMLString::compareString(attrname, SliceXMLString::sName))
      {
        name = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sValue))
      {
        value = attrval;
      }
    }

    handleSimpleOption(name, value);

    if (name)       delete [] name;
    if (value)      delete [] value;
  }
  else if (!XMLString::compareString(localname, SliceXMLString::sRequest))
  {
    // <Request> Information
   
    char * id = NULL;
    char * type = NULL;
    char * host = NULL;
    char * user = NULL;
    char * option = NULL;

    for (unsigned int i = 0; i < attrsLen; i++)
    {
      const XMLCh * attrname = attrs.getLocalName(i);
      char * attrval = XMLString::transcode(attrs.getValue(i));

      if (!XMLString::compareString(attrname, SliceXMLString::sType))
      {
        type = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sHost))
      {
        host = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sUser))
      {
        user = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sOption))
      {
        option = attrval;
      }
      else if (!XMLString::compareString(attrname, SliceXMLString::sId))
      {
        id = attrval;
      }
    }

    handleRequestData(type, host, user, id, option);

    if (id)     delete [] id;
    if (type)   delete [] type;
    if (host)   delete [] host;
    if (user)   delete [] user;
    if (option) delete [] option;
  }
  
  // Other elements need no action
}

//! Accepts character data from the XML parser
/*! Data that is not in a tag or its attributes <TAGNAME attribute=val ...> is
 *  passed to the handler with through this function by the XML API. It is
 *  the responsibility of the handler to make sense of it.\n\n
 *
 *  Fortunately, there  are only three cases where character data is valid to 
 *  be accepted, so the startElement function only needs to make note when the 
 *  XML engine is working in one of those three sections. Then the character 
 *  data is passed to the appropriate function in ConsensusRequest. \n\n
 *
 *  Also Xerces makes no guarantee that there will only be one call to 
 *  characters() per character data block so the accepting funtions need to 
 *  keep appending into an internal buffer.
 */
void SliceSAX2Handler::characters(const XMLCh* const chars,
                                  const unsigned int length)
{
  if (m_doLogging)
  {
    m_log << "characters (";

    switch (m_location)
    {
      case IN_NUC:        m_log << "IN_NUC";                    break;
      case IN_READID:     m_log << "IN_READID";                 break;
      case IN_QUALVAL:    m_log << "IN_QUALVAL";                break;

      case IN_LEFT_QUAL:  m_log << "IN_LEFT_QUAL";              break;
      case IN_RIGHT_QUAL: m_log << "IN_RIGHT_QUAL";             break;

      case IN_LEFT_NUC:   m_log << "IN_LEFT_NUC";               break;
      case IN_RIGHT_NUC:  m_log << "IN_RIGHT_NUC";              break;

      default:            m_log << "UNKNOWN!";
    };

    m_log << ") \"" << StrX(chars) << "\"" << endl;
  }

  m_characterCount += length;

  if (m_location != IN_UNKNOWN)
  {
    const char * charbuf = XMLString::transcode(chars);
    
    switch (m_location)
    {
      case IN_NUC:        handleNuc(charbuf);                   break;
      case IN_READID:     handleReadId(charbuf);                break;
      case IN_QUALVAL:    handleQualVal(charbuf);               break;

      case IN_LEFT_QUAL:  handleTrimmedQual(charbuf, IN_LEFT);  break;
      case IN_RIGHT_QUAL: handleTrimmedQual(charbuf, IN_RIGHT); break;

      case IN_LEFT_NUC:   handleTrimmedNuc(charbuf, IN_LEFT);   break;
      case IN_RIGHT_NUC:  handleTrimmedNuc(charbuf, IN_RIGHT);  break;

      default:
        throw Exception("Characters in unexpected section",
                        "SliceSAX2Handler::characters");
    };

    if (charbuf) delete [] charbuf;
  }
}

void SliceSAX2Handler::endDocument()
{
  m_log << "endDocument" << endl;
  endCurrentSlice();
}

void SliceSAX2Handler::handleTrimmedSeq(const char * position)
{
  if (position)
  {
    if (!strcmp(position, "left"))
    {
      m_location = IN_LEFT;
    }
    else if (!strcmp(position, "right"))
    {
      m_location = IN_RIGHT;
    }
    else
    {
      throw Exception("Unknown position " + (string) position,
                      "SliceSAX2Handler::handleTrimmedSeq");
    }
  }
  else
  {
    throw Exception("Position is null",
                    "SliceSAX2Handler::handleTrimmedSeq");
  }
}

void SliceSAX2Handler::handleTrimmedNuc(const char * buffer,
                                        XMLLocation location)
{
  if (!m_currentRead)
  {
    throw Exception("Can't handle Clr Data without a Read",
                    "SliceSAX2Handler::handleTrimmedNuc");
  }

  m_currentRead->addNucData(buffer, location==IN_LEFT);
}

void SliceSAX2Handler::handleTrimmedQual(const char * buffer,
                                         XMLLocation location)
{
  if (!m_currentRead)
  {
    throw Exception("Can't handle Clr Data without a Read",
                    "SliceSAX2Handler::handleTrimmedQual");
  }

  m_currentRead->addQualData(buffer, location==IN_LEFT);
}


void SliceSAX2Handler::handleRequestData(const char * type, 
                                         const char * host, 
                                         const char * user, 
                                         const char * id,
                                         const char * option)
{
  m_manager->addRequest(new ConsensusRequest(type, host, user, id, option));
}

void SliceSAX2Handler::handleReadData(const char * id, 
                                      const char * seqname,
                                      const char * dir, 
                                      const char * chem)
{
  if (m_currentRead)
  {
    m_manager->addReadData(m_currentRead);
  }

  m_currentRead = m_readManager.addReadData(id, seqname, dir, chem);
  m_readCount++;
}

void SliceSAX2Handler::handleSimpleOption(const char * name,
                                          const char * value)
{
  m_manager->addSimpleOption(name, value);
}

void SliceSAX2Handler::handleClrData(const char * left,
                                     const char * right)
{
  if (!m_currentRead)
  {
    throw Exception("Can't handle Clr Data without a Read",
                    "SliceSAX2Handler::handleClrData");
  }

  m_currentRead->addClrData(left, right);
}

//! Creates a new SliceData object and sets it to be the current slice
/*! Calls endCurrentSlice() first to close the previous slice.
 *  
 *  @see endCurrentSlice
 *  @see SliceData::SliceData
 */
void SliceSAX2Handler::handleSliceData(const char * index, 
                                       const char * gindex,
                                       const char * existing)
{
  endCurrentSlice();

  m_currentSlice = new SliceData(m_readManager, index, gindex, existing,
                                 m_sliceRangeIndex, m_sliceRangeGindex);
  m_sliceCount++;
}

void SliceSAX2Handler::handleSliceRange(const char * index, 
                                        const char * gindex)
{
  endCurrentSlice();
  m_location=IN_UNKNOWN;

  m_sliceRangeIndex  = (index)  ? atol(index)  : 0;
  m_sliceRangeGindex = (gindex) ? atol(gindex) : 0;

  if (m_currentRead)
  {
    m_manager->addReadData(m_currentRead);
  }

  m_manager->addSliceRange(m_sliceRangeIndex, m_sliceRangeGindex);

  m_sliceRangeCount++;
}


//! Finalizes a slice
void SliceSAX2Handler::endCurrentSlice()
{
  if (m_currentSlice)
  {
    try
    {
      m_manager->addSlice(m_currentSlice);
    }
    catch (const Exception & e)
    {
      m_manager->errorSliceHandler(m_currentSlice, e);
    }

    m_currentSlice = NULL;
  }
}

//! Appends Nuc data to the current working slice
/*! 
 *  @see SliceData::addNuc
 *  @throws Exception If there is no current working slice
 */
void SliceSAX2Handler::handleNuc(const char * nucData)
{
  if (!m_currentSlice)
  {
    throw Exception("Can't handle Nuc Data without a slice",
                        "SliceSAX2Handler::handleNuc");
  }

  m_currentSlice->addNuc(nucData);
}

//! Appends read id data to the current working slice
/*! 
 *  @see SliceData::addReadId
 *  @throws Exception If there is no current working slice
 */
void SliceSAX2Handler::handleReadId(const char * readId)
{
  if (!m_currentSlice)
  {
    throw Exception("Can't handle ReadId Data without a slice",
                        "SliceSAX2Handler::handleReadId");
  }

  m_currentSlice->addReadId(readId);
}

//! Appends quality value data to the current working slice
/*! 
 *  @see SliceData::addQualVal
 *  @throws Exception If there is no current working slice
 */
void SliceSAX2Handler::handleQualVal(const char * qualVal)
{
  if (!m_currentSlice)
  {
    throw Exception("Can't handle QualVal Data without a slice",
                        "SliceSAX2Handler::handleQualVal");
  }

  m_currentSlice->addQualVal(qualVal);
}


