#ifndef SLICEDATA_HH
#define SLICEDATA_HH 1

/*! @file
 *  @brief Defines interface for SliceData
 */

#include <string>
#include "Exceptions.hh"
#include "Slice.h"
#include "ReadData.hh"
#include "ConsensusData.hh"

/*! @defgroup cppinterface C++ Interface
 *  Components in the C++ interface.
 *
 *  Read information is stored in ReadData objects and managed by a single
 *  ReadManager. Slice information is stored in SliceData objects, and the
 *  results of the calculations are managed in the ConsensusData objects.
 */


/*! @ingroup cppinterface 
 *  @brief The entire C++ interface is in the libSlice namespace
 */
namespace libSlice
{
  class ConsensusData;

  /*! @ingroup cppinterface 
   *  @brief Stores information about a given slice
   *
   *  Each slice is responsible for calculating its own results of calling 
   *  libSlice. Error information is also stored within the slice, so that
   *  all available information about the error cause slice is stored in a 
   *  single place.
   */
  class SliceData
  {
  public:
    SliceData(const ReadManager & readManager,
              const char * index, 
              const char * gindex,
              const char * existing,
              long sliceRangeIndex=0,
              long sliceRangeGindex=0);

    virtual ~SliceData();

    void addNuc     (const char * nucData);
    void addQualVal (const char * qualVal);
    void addReadId  (const char * readId);


    void print(ostream & out, const string & prefix = "");
    void printError(ostream & out) const;
    void printXML(ostream & out, const string & prefix = "    ");

    int getErrorCode() const;
    int getErrorOperation() const;

    void setDistributionTable(double freqA, double freqC,
                              double freqG, double freqT,
                              double freqGap);

    void setHighQualityThreshold(int highQualityThreshold);
    void setDoAmbiguity(int doAmbiguity);
    void setScoringModel(int scoringModel);

    ConsensusData doCalculations(bool recallEmpty = false);

    //! Error Codes for slices
    enum SliceErrorCode
    {
      SLICE_NOERROR = 0,
      SLICE_ERR_GETCONSENSUS,
      SLICE_ERR_QV_NONNUMERIC,
      SLICE_ERR_UNDERFLOW,
      SLICE_ERR_OVERFLOW,
      SLICE_ERR_NO_READID,
      SLICE_ERR_CALCULATION_NOT_DONE
    };

    //! Special codes for string operations errrors
    enum StringOperation
    {
      STROP_NONE = 0,
      STROP_GET_QUAL_VAL,
      STROP_IS_REVERSE_COMPLIMENT
    };

    unsigned short   getDepth()                     const;
    ReadData *       getRead(unsigned short index)  const;
    long             getIndex()                     const;
    long             getGindex()                    const;

    char             getBase(unsigned short index);
    char             getQualVal(unsigned short index);
    char             getConsensus();

    static char      getBaseCompliment(char base);

    const string &   getNucData()                    const;
    const string &   getQualValData()                const;
    const string &   getReadIdData()                 const;

    void             resetIndex(long index, long gindex);
    void             resetReadId();
    void             resetQualVal();
    void             resetNuc();
    void             setConsensus(char consensus);


  private:
    const ReadManager & m_readManager;

    void           flattenSlice();

    char *         getBases(int len) const;

    char           getExisting()                    const;
    char *         getQualVals(int len);
    char *         getReverseCompliments(int len);
    char *         foreach(const string & data, int numresults,
                           StringOperation operation,
                           const char * underflowError,
                           const char * overflowError);

    libSlice_Slice * m_slice;

    long m_index;
    long m_gindex;
    string m_existing;
    string m_indexstr;

    string m_nucData;
    string m_qualValData;
    string m_readIdData;

    int    m_highQualityThreshold;
    int    m_doAmbiguity;
    int    m_scoringModel;

    libSlice_BaseDistribution m_distribution;

    SliceErrorCode  m_errorCode;
    StringOperation m_errorOperation;
    string          m_errorData;
  };
}

ostream & operator<< (ostream & os, libSlice::SliceData & slice);
ostream & operator<< (ostream & os, libSlice::SliceData * slice);


#endif


